"""Object Graph"""
from contextlib import contextmanager

from microcosm.configuration import Configuration
from microcosm.errors import CyclicGraphError, LockedGraphError
from microcosm.decorators import get_defaults
from microcosm.loaders import load_from_environ
from microcosm.metadata import Metadata
from microcosm.modules import ModuleFinder
from microcosm.registry import _registry


RESERVED = object()


class ObjectGraph(object):
    """
    An object graph contains all of the instantiated components for a microservice.

    Because components can reference each other acyclically, this collection of
    components forms a directed acyclic graph.

    """
    def __init__(self, metadata, config, registry):
        self.metadata = metadata
        self.config = config
        self._locked = False
        self._registry = registry
        self._components = {}

    def use(self, *keys):
        """
        Explicitly initialize a set of components by their binding keys.

        """
        return [
            getattr(self, key)
            for key in keys
        ]

    def lock(self):
        """
        Lock the graph so that new components cannot be created.

        """
        self._locked = True

    def unlock(self):
        """
        Unlock the graph so that new components can created.

        """
        self._locked = False

    def export(self, package_name):
        """
        Export this object graph under the given package name.

        Allows the graph's components to be accessible via an import statement thereby
        making an object graph usable by code paths that don't have easy access to the
        graph.

        Usage:

            from <package_name> import <component>

        To avoid inconsistencies between the graph and the module, the graph will be locked
        by this operation.

        """
        self.lock()
        ModuleFinder.export(self, package_name)

    def __getattr__(self, key):
        """
        Access a component by its binding key.

        If the component is not present, it will be lazily created.

        :raises CyclicGraphError: if the factory function requires a cycle
        :raises LockedGraphError: if the graph is locked

        """
        try:
            component = self._components[key]
            if component is RESERVED:
                raise CyclicGraphError()
            return component
        except KeyError:
            if self._locked:
                raise LockedGraphError()
            return self._resolve_key(key)

    @contextmanager
    def _reserve(self, key):
        """
        Reserve a component's binding temporarily.

        Protects against cycles.

        """
        self._components[key] = RESERVED
        try:
            yield
        finally:
            del self._components[key]

    def _resolve_key(self, key):
        """
        Attempt to lazily create a component.

        :raises NotBoundError: if the component does not have a bound factory
        :raises CyclicGraphError: if the factory function requires a cycle
        :raises LockedGraphError: if the graph is locked
        """
        with self._reserve(key):
            factory = self._registry.resolve(key)
            component = factory(self)

        self._components[key] = component
        return component


def create_object_graph(name,
                        debug=False,
                        testing=False,
                        import_name=None,
                        root_path=None,
                        loader=load_from_environ,
                        registry=_registry):
    """
    Create a new object graph.

    :param name: the name of the microservice
    :param debug: is development debugging enabled?
    :param testing: is unit testing enabled?
    :param loader: the configuration loader to use
    :param registry: the registry to use (defaults to the global)
    """
    metadata = Metadata(
        name=name,
        debug=debug,
        testing=testing,
        import_name=import_name,
        root_path=root_path,
    )

    config = Configuration({
        key: get_defaults(value)
        for key, value in registry.all.items()
    })
    config.merge(loader(metadata))

    return ObjectGraph(
        metadata=metadata,
        config=config,
        registry=registry,
    )



"""Factory decorators"""
from microcosm.registry import _registry


DEFAULTS = "_defaults"


def binding(key, registry=_registry):
    """
    Creates a decorator that binds a factory function to a key.

    :param key: the binding key
    :param: registry: the registry to bind to; defaults to the global registry

    """
    def decorator(func):
        registry.bind(key, func)
        return func
    return decorator


def defaults(**kwargs):
    """
    Creates a decorator that saves the provided kwargs as defaults for a factory function.

    """
    def decorator(func):
        setattr(func, DEFAULTS, kwargs)
        return func
    return decorator


def get_defaults(func):
    """
    Retrieve the defaults for a factory function.

    """
    return getattr(func, DEFAULTS, {})



"""Error types"""


class AlreadyBoundError(Exception):
    """
    Raised if a factory is already bound to a name.

    """
    pass


class CyclicGraphError(Exception):
    """
    Raised if a graph has a cycle.

    """
    pass


class LockedGraphError(Exception):
    """
    Raised when attempting to create a component in a locked object graph.

    """
    pass


class NotBoundError(Exception):
    """
    Raised if not factory is bound to a name.

    """
    pass



"""Example component factory"""


def create_hello_world(graph):
    """
    Produces the string "hello world"

    Provided as demonstration and test bed for entrypoint-based registration.

    """
    return "hello world"



"""Service metadata"""
from os.path import abspath, dirname, join
from sys import modules


class Metadata(object):
    """
    Service metadata.

    Factories should use service metadata to implement conventions.

    """

    def __init__(self, name, debug=False, testing=False, import_name=None, root_path=None):
        """
        :param name: the name of the microservice
        :param debug: is development debugging enabled?
        :param testing: is unit testing enabled?
        :param import_name: the import name to use for resource loading
        :param root_path: the root path for resource loading

        """
        self.name = name
        self.debug = debug
        self.testing = testing
        self.root_path = root_path or self.get_root_path(import_name or name)

    def get_root_path(self, name):
        """
        Attempt to compute a root path for a (hopefully importable) name.

        Based in part on Flask's `root_path` calculation. See:

            https://github.com/mitsuhiko/flask/blob/master/flask/helpers.py#L777

        """
        module = modules.get(name)
        if module is not None and hasattr(module, '__file__'):
            return dirname(abspath(module.__file__))

        # Flask keeps looking at this point. We instead set the root path to None,
        # assume that the user doesn't need resource loading, and raise an error
        # when resolving the resource path.
        return None

    def get_path(self, path):
        if self.root_path is None:
            raise RuntimeError("Root path was not defined. Either use a resolvable name or set root path explicitly.")
        return join(self.root_path, path)



"""
Allows object graph members to be imported by defining custom finders and loaders.

"""
from imp import new_module
import sys


class ModuleFinder(object):
    """
    A module "finder" that matches a configured package name.
    """
    def __init__(self, graph, package_name):
        """
        :param graph: the object graph
        :param package_name: the package name
        """
        self.graph = graph
        self.package_name = package_name

    def find_module(self, fullname, path=None):
        if self.package_name.startswith(fullname):
            return ModuleLoader(self.graph, self.package_name)
        else:
            return None

    @classmethod
    def export(cls, graph, package_name):
        module_finder = cls(
            graph,
            package_name,
        )
        sys.meta_path.append(module_finder)


class ModuleLoader(object):
    """
    A module "loader" that returns object graph components.
    """
    def __init__(self, graph, package_name):
        """
        :param graph: the object graph
        :param package_name: the package name prefix
        """
        self.graph = graph
        self.package_name = package_name

    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]

        module = self.make_module(fullname)
        sys.modules[fullname] = module

        if fullname == self.package_name:
            module.__dict__.update(self.graph._components)

        return module

    def make_module(self, fullname):
        module = new_module(fullname)
        module.__file__ = "<microcosm>"
        module.__loader__ = self
        module.__package__ = fullname
        module.__path__ = []
        return module



"""Configuration model and loading"""


class Configuration(dict):
    """
    Nested attribute dictionary with recursive merging for modeling configuration.

    Based on the `easydict` project, but locally modified to remove class attribute
    support (which interferes with adding new member functions). Note that some
    dict functions (`update`, `pop`) are not correctly implemented, but are also not
    needed here.

    """
    def __init__(self, dct=None, **kwargs):
        if dct is None:
            dct = {}
        if kwargs:
            dct.update(**kwargs)

        for key, value in dct.items():
            setattr(self, key, value)

    def __setattr__(self, name, value):
        if isinstance(value, (list, tuple)):
            value = [
                self.__class__(x)
                if isinstance(x, dict) else x for x in value
            ]
        else:
            value = self.__class__(value) if isinstance(value, dict) else value
        super(Configuration, self).__setattr__(name, value)
        super(Configuration, self).__setitem__(name, value)

    def merge(self, dct=None, **kwargs):
        """
        Recursively merge a dictionary or kwargs into the current dict.

        """
        if dct is None:
            dct = {}
        if kwargs:
            dct.update(**kwargs)

        for key, value in dct.items():
            if isinstance(value, dict) and isinstance(self[key], Configuration):
                # recursively merge
                self[key].merge(value)
            else:
                # set the new value
                self[key] = value

    __setitem__ = __setattr__



"""
Configuration loading

A configuration loader is any function that accepts `Metadata` and
returns a `Configuration` object. Configuration might be loaded
from a file, from environment variables, or from an external service.
"""
from imp import new_module
from json import loads
from os import environ

from inflection import underscore

from microcosm.configuration import Configuration


def get_config_filename(metadata):
    """
    Derive a configuration file name from the FOO_SETTINGS
    environment variable.

    """
    envvar = "{}_SETTINGS".format(underscore(metadata.name).upper())
    try:
        return environ[envvar]
    except KeyError:
        return None


def _load_from_file(metadata, load_func):
    """
    Load configuration from a file.

    The file path is derived from an environment variable
    named after the service of the form FOO_SETTINGS.

    """
    config_filename = get_config_filename(metadata)
    if config_filename is None:
        return Configuration()

    with open(config_filename, "r") as file_:
        data = load_func(file_.read())
        return Configuration(data)


def load_from_json_file(metadata):
    """
    Load configuration from a JSON file.

    """
    return _load_from_file(metadata, loads)


def load_from_python_file(metadata):
    """
    Load configuration from a Python file.

    The file path is derived from an environment variable
    named after the service of the form FOO_SETTINGS.

    """
    def load_python_module(data):
        module = new_module("magic")
        exec data in module.__dict__, module.__dict__
        return {
            key: value
            for key, value in module.__dict__.items()
            if not key.startswith("_")
        }
    return _load_from_file(metadata, load_python_module)


def load_from_environ(metadata):
    """
    Load configuration from environment variables.

    Any environment variable prefixed with the metadata's name will be
    used to recursively set dictionary keys, splitting on '_'.

    """
    # We'll match the ennvar name against the metadata's name. The ennvar
    # name must be uppercase and hyphens in names converted to underscores.
    #
    # | envar       | name    | matches? |
    # +-------------+---------+----------+
    # | FOO_BAR     | foo     | yes      |
    # | FOO_BAR     | bar     | no       |
    # | foo_bar     | bar     | no       |
    # | FOO_BAR_BAZ | foo_bar | yes      |
    # | FOO_BAR_BAZ | foo-bar | yes      |
    # +-------------+---------+----------+

    prefix = metadata.name.upper().replace("-", "_").split("_")

    def matches_key(key_parts):
        return len(key_parts) > len(prefix) and key_parts[:len(prefix)] == prefix

    config = Configuration()
    for key, value in environ.items():
        key_parts = key.split("_")
        if not matches_key(key_parts):
            continue

        dct = config
        # for each part before the last
        for key_part in key_parts[len(prefix):-1]:
            # build up the nested dictionary structure
            dct[key_part.lower()] = dict()
            dct = dct[key_part.lower()]
        # set the value for the final part
        dct[key_parts[-1].lower()] = value
    return config


def load_each(*loaders):
    """
    Loader factory that combines a series of loaders.

    """
    def _load_each(metadata):
        config = loaders[0](metadata)
        for loader in loaders[1:]:
            next_config = loader(metadata)
            config.merge(next_config)
        return config
    return _load_each



"""Registry of component factories"""
from itertools import chain
from pkg_resources import iter_entry_points

from lazy import lazy

from microcosm.errors import AlreadyBoundError, NotBoundError


class Registry(object):
    """
    Registry of component factories.

    Supports factories resolved explicitly and via entrypoints.

    """
    def __init__(self):
        self.factories = {}

    @lazy
    def entry_points(self):
        return {
            entry_point.name: entry_point.load()
            # NB: it's possible to have two entry points for the same name
            # (but in different distributions). This will cause unpredictable
            # behavior; don't do that.
            for entry_point in iter_entry_points(group="microcosm.factories")
        }

    @property
    def all(self):
        """
        Return a synthetic dictionary of all factories.
        """
        return {
            key: value
            for key, value in chain(self.entry_points.items(), self.factories.items())
        }

    def bind(self, key, factory):
        """
        Bind a factory to a key.

        :raises AlreadyBoundError: if the key is alrady bound

        """
        if key in self.factories:
            raise AlreadyBoundError(key)
        else:
            self.factories[key] = factory

    def resolve(self, key):
        """
        Resolve a key to a factory.

        Attempts to resolve explicit bindings and entry points, preferring
        explicit bindings.

        :raises NotBoundError: if the key cannot be resolved

        """
        try:
            return self._resolve_from_binding(key)
        except NotBoundError:
            return self._resolve_from_entry_point(key)

    def _resolve_from_binding(self, key):
        """
        Resolve using bindings.

        """
        try:
            return self.factories[key]
        except KeyError:
            raise NotBoundError(key)

    def _resolve_from_entry_point(self, key):
        """
        Resolve using entry points.

        """
        try:
            return self.entry_points[key]
        except KeyError:
            raise NotBoundError(key)


# global registry
_registry = Registry()
